@@ -53,10 +53,12 @@ except ImportError:
     import pickle
 
 __author__    = "Evan Martin <martine@danga.com>"
-__version__ = "1.31"
+__version__ = "1.34"
 __copyright__ = "Copyright (C) 2003 Danga Interactive"
 __license__   = "Python"
 
+SERVER_MAX_KEY_LENGTH = 250
+
 class _Error(Exception):
     pass
 
@@ -87,6 +89,14 @@ class Client:
 
     _SERVER_RETRIES = 10  # how many times to try finding a free server.
 
+    # exceptions for Client
+    class MemcachedKeyError(Exception):
+      pass
+    class MemcachedKeyLengthError(MemcachedKeyError):
+      pass
+    class MemcachedKeyCharacterError(MemcachedKeyError):
+      pass
+
     def __init__(self, servers, debug=0):
         """
         Create a new Client object with the given list of servers.
@@ -190,6 +200,7 @@ class Client:
         @return: Nonzero on success.
         @rtype: int
         '''
+        check_key(key)
         server, key = self._get_server(key)
         if not server:
             return 0
@@ -245,6 +256,7 @@ class Client:
         return self._incrdecr("decr", key, delta)
 
     def _incrdecr(self, cmd, key, delta):
+        check_key(key)
         server, key = self._get_server(key)
         if not server:
             return 0
@@ -293,6 +305,7 @@ class Client:
         return self._set("set", key, val, time)
     
     def _set(self, cmd, key, val, time):
+        check_key(key)
         server, key = self._get_server(key)
         if not server:
             return 0
@@ -326,6 +339,7 @@ class Client:
         
         @return: The value or None.
         '''
+        check_key(key)
         server, key = self._get_server(key)
         if not server:
             return None
@@ -371,6 +385,7 @@ class Client:
 
         # build up a list for each server of all the keys we want.
         for key in keys:
+            check_key(key)
             server, key = self._get_server(key)
             if not server:
                 continue
@@ -467,6 +482,8 @@ class _Host:
         self.deaduntil = 0
         self.socket = None
 
+        self.buffer = ''
+
     def _check_dead(self):
         if self.deaduntil and self.deaduntil > time.time():
             return 1
@@ -496,6 +513,7 @@ class _Host:
             self.mark_dead("connect: %s" % msg[1])
             return None
         self.socket = s
+        self.buffer = ''
         return s
     
     def close_socket(self):
@@ -507,18 +525,24 @@ class _Host:
         self.socket.sendall(cmd + '\r\n')
 
     def readline(self):
-        buffers = ''
+        buf = self.buffer
         recv = self.socket.recv
-        while 1:
-            data = recv(1)
+        while True:
+            index = buf.find('\r\n')
+            if index >= 0:
+                break
+            data = recv(4096)
             if not data:
                 self.mark_dead('Connection closed while reading from %s'
                         % repr(self))
                 break
-            if data == '\n' and buffers and buffers[-1] == '\r':
-                return(buffers[:-1])
-            buffers = buffers + data
-        return(buffers)
+            buf += data
+        if index >= 0:
+            self.buffer = buf[index+2:]
+            buf = buf[:index]
+        else:
+            self.buffer = ''
+        return buf
 
     def expect(self, text):
         line = self.readline()
@@ -527,11 +551,16 @@ class _Host:
         return line
     
     def recv(self, rlen):
-        buf = ''
-        recv = self.socket.recv
+        self_socket_recv = self.socket.recv
+        buf = self.buffer
         while len(buf) < rlen:
-            buf = buf + recv(rlen - len(buf))
-        return buf
+            foo = self_socket_recv(4096)
+            buf += foo
+            if len(foo) == 0:
+                raise _Error, ( 'Read %d bytes, expecting %d, '
+                        'read returned 0 length bytes' % ( len(buf), foo ))
+        self.buffer = buf[rlen:]
+        return buf[:rlen]
 
     def __str__(self):
         d = ''
@@ -539,6 +568,17 @@ class _Host:
             d = " (dead until %d)" % self.deaduntil
         return "%s:%d%s" % (self.ip, self.port, d)
 
+def check_key(key):
+    """
+    Checks to make sure the key is less than SERVER_MAX_KEY_LENGTH or contains control characters.
+    If test fails throws MemcachedKeyLength error.
+    """
+    if len(key) > SERVER_MAX_KEY_LENGTH:
+      raise Client.MemcachedKeyLengthError, "Key length is > %s" % SERVER_MAX_KEY_LENGTH
+    for char in key:
+      if ord(char) < 33:
+        raise Client.MemcachedKeyCharacterError, "Control characters not allowed"
+
 def _doctest():
     import doctest, memcache
     servers = ["127.0.0.1:11211"]
